@@ -257,14 +257,17 @@ class Application:
     def planned_units(self) -> int:
         """Get the number of units that Juju has "planned" for this application.
 
-        E.g., if an operator ran "juju deploy foo", then "juju add-unit -n 2 foo", the
-        planned unit count for foo would be 3.
+        E.g., if an operator runs "juju deploy foo", then "juju add-unit -n 2 foo", the
+        planned unit count for foo will be 3.
 
-        We deliberately do not attempt to inspect whether these units are actually running
-        or not at present.
+        The data comes from the Juju agent, based on data it fetches from the
+        controller. Pending units are included in the count, and scale down events may
+        modify the count before some units have been fully torn down. The information in
+        planned_units is up-to-date as of the start of the current hook invocation.
+
+        This method only returns data for this charm's application -- the Juju agent isn't
+        able to see planned unit counts for other applications in the model.
 
-        This only works for this charm's application -- the unit agent isn't able to get
-        planned units for other applications in the model.
         """
         if not self._is_our_app:
             raise RuntimeError(
@@ -1194,13 +1197,14 @@ class Container:
         If no service names are specified, return status information for all
         services, otherwise return information for only the given services.
         """
+        service_names = service_names or None
         services = self._pebble.get_services(service_names)
         return ServiceInfoMapping(services)
 
     def get_service(self, service_name: str) -> 'pebble.ServiceInfo':
         """Get status information for a single named service.
 
-        Raises model error if service_name is not found.
+        Raises :class:`ModelError` if service_name is not found.
         """
         services = self.get_services(service_name)
         if not services:
@@ -1209,6 +1213,33 @@ class Container:
             raise RuntimeError('expected 1 service, got {}'.format(len(services)))
         return services[service_name]
 
+    def get_checks(
+            self,
+            *check_names: str,
+            level: 'pebble.CheckLevel' = None) -> 'CheckInfoMapping':
+        """Fetch and return a mapping of check information indexed by check name.
+
+        Args:
+            check_names: Optional check names to query for. If no check names
+                are specified, return checks with any name.
+            level: Optional check level to query for. If not specified, fetch
+                checks with any level.
+        """
+        checks = self._pebble.get_checks(names=check_names or None, level=level)
+        return CheckInfoMapping(checks)
+
+    def get_check(self, check_name: str) -> 'pebble.CheckInfo':
+        """Get check information for a single named check.
+
+        Raises :class:`ModelError` if check_name is not found.
+        """
+        checks = self.get_checks(check_name)
+        if not checks:
+            raise ModelError('check {!r} not found'.format(check_name))
+        if len(checks) > 1:
+            raise RuntimeError('expected 1 check, got {}'.format(len(checks)))
+        return checks[check_name]
+
     def pull(self, path: str, *, encoding: str = 'utf-8') -> typing.Union[typing.BinaryIO,
                                                                           typing.TextIO]:
         """Read a file's content from the remote system.
@@ -1269,6 +1300,26 @@ class Container:
         """
         return self._pebble.list_files(path, pattern=pattern, itself=itself)
 
+    def exists(self, path: str) -> bool:
+        """Return true if the path exists on the container filesystem."""
+        try:
+            self._pebble.list_files(path, itself=True)
+        except pebble.APIError as err:
+            if err.code == 404:
+                return False
+            raise err
+        return True
+
+    def isdir(self, path: str) -> bool:
+        """Return true if a directory exists at the given path on the container filesystem."""
+        try:
+            files = self._pebble.list_files(path, itself=True)
+        except pebble.APIError as err:
+            if err.code == 404:
+                return False
+            raise err
+        return files[0].type == pebble.FileType.DIRECTORY
+
     def make_dir(
             self, path: str, *, make_parents: bool = False, permissions: int = None,
             user_id: int = None, user: str = None, group_id: int = None, group: str = None):
@@ -1378,7 +1429,7 @@ class ContainerMapping(Mapping):
 
 
 class ServiceInfoMapping(Mapping):
-    """Map of service names to pebble.ServiceInfo objects.
+    """Map of service names to :class:`ops.pebble.ServiceInfo` objects.
 
     This is done as a mapping object rather than a plain dictionary so that we
     can extend it later, and so it's not mutable.
@@ -1400,6 +1451,29 @@ class ServiceInfoMapping(Mapping):
         return repr(self._services)
 
 
+class CheckInfoMapping(Mapping):
+    """Map of check names to :class:`ops.pebble.CheckInfo` objects.
+
+    This is done as a mapping object rather than a plain dictionary so that we
+    can extend it later, and so it's not mutable.
+    """
+
+    def __init__(self, checks: typing.Iterable['pebble.CheckInfo']):
+        self._checks = {c.name: c for c in checks}
+
+    def __getitem__(self, key: str):
+        return self._checks[key]
+
+    def __iter__(self):
+        return iter(self._checks)
+
+    def __len__(self):
+        return len(self._checks)
+
+    def __repr__(self):
+        return repr(self._checks)
+
+
 class ModelError(Exception):
     """Base class for exceptions raised when interacting with the Model."""
     pass
