@@ -19,6 +19,7 @@ For a command-line interface for local testing, see test/pebble_cli.py.
 
 import binascii
 import cgi
+import copy
 import datetime
 import email.parser
 import enum
@@ -33,8 +34,10 @@ import shutil
 import signal
 import socket
 import sys
+import tempfile
 import threading
 import time
+import types
 import typing
 import urllib.error
 import urllib.parse
@@ -511,13 +514,19 @@ class Change:
 
 
 class Plan:
-    """Represents the effective Pebble configuration."""
+    """Represents the effective Pebble configuration.
+
+    A plan is the combined layer configuration. The layer configuration is
+    documented at https://github.com/canonical/pebble/#layer-specification.
+    """
 
     def __init__(self, raw: str):
         d = yaml.safe_load(raw) or {}
         self._raw = raw
         self._services = {name: Service(name, service)
                           for name, service in d.get('services', {}).items()}
+        self._checks = {name: Check(name, check)
+                        for name, check in d.get('checks', {}).items()}
 
     @property
     def services(self):
@@ -527,14 +536,21 @@ class Plan:
         """
         return self._services
 
+    @property
+    def checks(self):
+        """This plan's checks mapping (maps check name to :class:`Check`).
+
+        This property is currently read-only.
+        """
+        return self._checks
+
     def to_dict(self) -> typing.Dict[str, typing.Any]:
         """Convert this plan to its dict representation."""
-        as_dicts = {name: service.to_dict() for name, service in self._services.items()}
-        if not as_dicts:
-            return {}
-        return {
-            'services': as_dicts,
-        }
+        fields = [
+            ('services', {name: service.to_dict() for name, service in self._services.items()}),
+            ('checks', {name: check.to_dict() for name, check in self._checks.items()}),
+        ]
+        return {name: value for name, value in fields if value}
 
     def to_yaml(self) -> str:
         """Return this plan's YAML representation."""
@@ -546,13 +562,14 @@ class Plan:
 class Layer:
     """Represents a Pebble configuration layer.
 
-    The format of this is not documented, but is captured in code here:
-    https://github.com/canonical/pebble/blob/master/internal/plan/plan.go
+    The format of this is documented at
+    https://github.com/canonical/pebble/#layer-specification.
 
     Attributes:
         summary: A summary of the purpose of this layer
         description: A long form description of this layer
-        services: A mapping of name: :class:`Service` defined by this layer
+        services: A mapping of name to :class:`Service` defined by this layer
+        checks: A mapping of check to :class:`Check` defined by this layer
     """
 
     # This is how you do type annotations, but it is not supported by Python 3.5
@@ -569,6 +586,8 @@ class Layer:
         self.description = d.get('description', '')
         self.services = {name: Service(name, service)
                          for name, service in d.get('services', {}).items()}
+        self.checks = {name: Check(name, check)
+                       for name, check in d.get('checks', {}).items()}
 
     def to_yaml(self) -> str:
         """Convert this layer to its YAML representation."""
@@ -579,7 +598,8 @@ class Layer:
         fields = [
             ('summary', self.summary),
             ('description', self.description),
-            ('services', {name: service.to_dict() for name, service in self.services.items()})
+            ('services', {name: service.to_dict() for name, service in self.services.items()}),
+            ('checks', {name: check.to_dict() for name, check in self.checks.items()}),
         ]
         return {name: value for name, value in fields if value}
 
@@ -608,8 +628,14 @@ class Service:
         self.user_id = raw.get('user-id')
         self.group = raw.get('group', '')
         self.group_id = raw.get('group-id')
+        self.on_success = raw.get('on-success', '')
+        self.on_failure = raw.get('on-failure', '')
+        self.on_check_failure = dict(raw.get('on-check-failure', {}))
+        self.backoff_delay = raw.get('backoff-delay', '')
+        self.backoff_factor = raw.get('backoff-factor')
+        self.backoff_limit = raw.get('backoff-limit', '')
 
-    def to_dict(self) -> typing.Dict:
+    def to_dict(self) -> typing.Dict[str, typing.Any]:
         """Convert this service object to its dict representation."""
         fields = [
             ('summary', self.summary),
@@ -625,6 +651,12 @@ class Service:
             ('user-id', self.user_id),
             ('group', self.group),
             ('group-id', self.group_id),
+            ('on-success', self.on_success),
+            ('on-failure', self.on_failure),
+            ('on-check-failure', self.on_check_failure),
+            ('backoff-delay', self.backoff_delay),
+            ('backoff-factor', self.backoff_factor),
+            ('backoff-limit', self.backoff_limit),
         ]
         return {name: value for name, value in fields if value}
 
@@ -700,6 +732,78 @@ class ServiceInfo:
                 ).format(self=self)
 
 
+class Check:
+    """Represents a check in a Pebble configuration layer."""
+
+    def __init__(self, name: str, raw: typing.Dict = None):
+        self.name = name
+        raw = raw or {}
+        self.override = raw.get('override', '')
+        try:
+            self.level = CheckLevel(raw.get('level', ''))
+        except ValueError:
+            self.level = raw.get('level')
+        self.period = raw.get('period', '')
+        self.timeout = raw.get('timeout', '')
+        self.threshold = raw.get('threshold')
+
+        http = raw.get('http')
+        if http is not None:
+            http = copy.deepcopy(http)
+        self.http = http
+
+        tcp = raw.get('tcp')
+        if tcp is not None:
+            tcp = copy.deepcopy(tcp)
+        self.tcp = tcp
+
+        exec_ = raw.get('exec')
+        if exec_ is not None:
+            exec_ = copy.deepcopy(exec_)
+        self.exec = exec_
+
+    def to_dict(self) -> typing.Dict[str, typing.Any]:
+        """Convert this check object to its dict representation."""
+        fields = [
+            ('override', self.override),
+            ('level', self.level.value),
+            ('period', self.period),
+            ('timeout', self.timeout),
+            ('threshold', self.threshold),
+            ('http', self.http),
+            ('tcp', self.tcp),
+            ('exec', self.exec),
+        ]
+        return {name: value for name, value in fields if value}
+
+    def __repr__(self) -> str:
+        return 'Check({!r})'.format(self.to_dict())
+
+    def __eq__(self, other: typing.Union[typing.Dict, 'Check']) -> bool:
+        """Compare this check configuration to another."""
+        if isinstance(other, dict):
+            return self.to_dict() == other
+        elif isinstance(other, Check):
+            return self.to_dict() == other.to_dict()
+        else:
+            raise ValueError("Cannot compare pebble.Check to {}".format(type(other)))
+
+
+class CheckLevel(enum.Enum):
+    """Enum of check levels."""
+
+    UNSET = ''
+    ALIVE = 'alive'
+    READY = 'ready'
+
+
+class CheckStatus(enum.Enum):
+    """Enum of check statuses."""
+
+    UP = 'up'
+    DOWN = 'down'
+
+
 class FileType(enum.Enum):
     """Enum of file types."""
 
@@ -774,6 +878,68 @@ class FileInfo:
                 ).format(self=self)
 
 
+class CheckInfo:
+    """Check status information.
+
+    A list of these objects is returned from :meth:`Client.get_checks`.
+
+    Attributes:
+        name: The name of the check.
+        level: The check level: :attr:`CheckLevel.ALIVE`,
+            :attr:`CheckLevel.READY`, or None (level not set).
+        status: The status of the check: :attr:`CheckStatus.UP` means the
+            check is healthy (the number of failures is less than the
+            threshold), :attr:`CheckStatus.DOWN` means the check is unhealthy
+            (the number of failures has reached the threshold).
+        failures: The number of failures since the check last succeeded (reset
+            to zero if the check succeeds).
+        threshold: The failure threshold, that is, how many consecutive
+            failures for the check to be considered "down".
+    """
+
+    def __init__(
+        self,
+        name: str,
+        level: typing.Optional[typing.Union[CheckLevel, str]],
+        status: typing.Union[CheckStatus, str],
+        failures: int = 0,
+        threshold: int = 0,
+    ):
+        self.name = name
+        self.level = level
+        self.status = status
+        self.failures = failures
+        self.threshold = threshold
+
+    @classmethod
+    def from_dict(cls, d: typing.Dict) -> 'CheckInfo':
+        """Create new :class:`CheckInfo` object from dict parsed from JSON."""
+        try:
+            level = CheckLevel(d.get('level', ''))
+        except ValueError:
+            level = d.get('level')
+        try:
+            status = CheckStatus(d['status'])
+        except ValueError:
+            status = d['status']
+        return cls(
+            name=d['name'],
+            level=level,
+            status=status,
+            failures=d.get('failures', 0),
+            threshold=d['threshold'],
+        )
+
+    def __repr__(self):
+        return ('CheckInfo('
+                'name={self.name!r}, '
+                'level={self.level!r}, '
+                'status={self.status}, '
+                'failures={self.failures}, '
+                'threshold={self.threshold!r})'
+                ).format(self=self)
+
+
 class ExecProcess:
     """Represents a process started by :meth:`Client.exec`.
 
@@ -1069,6 +1235,8 @@ class _WebsocketReader(io.BufferedIOBase):
 class Client:
     """Pebble API client."""
 
+    _chunk_size = 8192
+
     def __init__(self, socket_path=None, opener=None, base_url='http://localhost', timeout=5.0):
         """Initialize a client instance.
 
@@ -1132,7 +1300,12 @@ class Client:
         """Make a request to the Pebble server; return the raw HTTPResponse object."""
         url = self.base_url + path
         if query:
-            url = url + '?' + urllib.parse.urlencode(query)
+            url = url + '?' + urllib.parse.urlencode(query, doseq=True)
+
+        # python 3.5 urllib requests require their data to be a bytes object -
+        # generators won't work.
+        if sys.version_info[:2] < (3, 6) and isinstance(data, types.GeneratorType):
+            data = b''.join(data)
 
         if headers is None:
             headers = {}
@@ -1422,7 +1595,7 @@ class Client:
         self._request('POST', '/v1/layers', body=body)
 
     def get_plan(self) -> Plan:
-        """Get the Pebble plan (currently contains only combined services)."""
+        """Get the Pebble plan (contains combined layer configuration)."""
         resp = self._request('GET', '/v1/plan', {'format': 'yaml'})
         return Plan(resp['result'])
 
@@ -1464,45 +1637,40 @@ class Client:
         if not boundary:
             raise ProtocolError('invalid boundary {!r}'.format(boundary))
 
-        # We have to manually write the Content-Type with boundary, because
-        # email.parser expects the entire multipart message with headers.
-        parser = email.parser.BytesFeedParser()
-        parser.feed(b'Content-Type: multipart/form-data; boundary='
-                    + boundary.encode('utf-8') + b'\r\n\r\n')
+        parser = _FilesParser(boundary)
 
-        # Then read the rest of the response and feed it to the parser.
         while True:
-            chunk = response.read(8192)
+            chunk = response.read(self._chunk_size)
             if not chunk:
                 break
             parser.feed(chunk)
-        message = parser.close()
-
-        # Walk over the multipart parts and read content and metadata.
-        resp = None
-        content = None
-        for part in message.walk():
-            name = part.get_param('name', header='Content-Disposition')
-            if name == 'response':
-                resp = _json_loads(part.get_payload())
-            elif name == 'files':
-                filename = part.get_filename()
-                if filename != path:
-                    raise ProtocolError('path not expected: {}'.format(filename))
-                # decode=True, ironically, avoids decoding bytes to str
-                content = part.get_payload(decode=True)
 
+        resp = parser.get_response()
         if resp is None:
             raise ProtocolError('no "response" field in multipart body')
         self._raise_on_path_error(resp, path)
 
-        if content is None:
+        filenames = parser.filenames()
+        if not filenames:
             raise ProtocolError('no file content in multipart response')
-        if encoding is not None:
-            reader = io.StringIO(content.decode(encoding))
-        else:
-            reader = io.BytesIO(content)
-        return reader
+        elif len(filenames) > 1:
+            raise ProtocolError('single file request resulted in a multi-file response')
+
+        filename = filenames[0]
+        if filename != path:
+            raise ProtocolError('path not expected: {!r}'.format(filename))
+
+        f = parser.get_file(path, encoding)
+
+        # The opened file remains usable after removing/unlinking on posix
+        # platforms until it is closed.  This prevents callers from
+        # interacting with it on disk.  But e.g. Windows doesn't allow
+        # removing opened files, and so we use the tempfile lib's
+        # helper class to auto-delete on close/gc for us.
+        if os.name != 'posix' or sys.platform == 'cygwin':
+            return tempfile._TemporaryFileWrapper(f, f.name, delete=True)
+        parser.remove_files()
+        return f
 
     @staticmethod
     def _raise_on_path_error(resp, path):
@@ -1546,14 +1714,7 @@ class Client:
             'files': [info],
         }
 
-        if hasattr(source, 'read'):
-            content = source.read()
-        else:
-            content = source
-        if isinstance(content, str):
-            content = content.encode(encoding)
-
-        data, content_type = self._encode_multipart(metadata, path, content)
+        data, content_type = self._encode_multipart(metadata, path, source, encoding)
 
         headers = {
             'Accept': 'application/json',
@@ -1579,28 +1740,46 @@ class Client:
             d['group'] = group
         return d
 
-    @staticmethod
-    def _encode_multipart(metadata, path, content):
+    def _encode_multipart(self, metadata, path, source, encoding):
         # Python's stdlib mime/multipart handling is screwy and doesn't handle
         # binary properly, so roll our own.
+
+        if isinstance(source, str):
+            source = io.StringIO(source)
+        elif isinstance(source, bytes):
+            source = io.BytesIO(source)
+
         boundary = binascii.hexlify(os.urandom(16))
         path_escaped = path.replace('"', '\\"').encode('utf-8')  # NOQA: test_quote_backslashes
-        parts = []
-        parts.extend([
-            b'--', boundary, b'\r\n',
-            b'Content-Type: application/json\r\n',
-            b'Content-Disposition: form-data; name="request"\r\n',
-            b'\r\n',
-            json.dumps(metadata).encode('utf-8'), b'\r\n',
-            b'--', boundary, b'\r\n',
-            b'Content-Type: application/octet-stream\r\n',
-            b'Content-Disposition: form-data; name="files"; filename="', path_escaped, b'"\r\n',
-            b'\r\n',
-            content, b'\r\n',
-            b'--', boundary, b'--\r\n',
-        ])
         content_type = 'multipart/form-data; boundary="' + boundary.decode('utf-8') + '"'
-        return b''.join(parts), content_type
+
+        def generator():
+            yield b''.join([
+                b'--', boundary, b'\r\n',
+                b'Content-Type: application/json\r\n',
+                b'Content-Disposition: form-data; name="request"\r\n',
+                b'\r\n',
+                json.dumps(metadata).encode('utf-8'), b'\r\n',
+                b'--', boundary, b'\r\n',
+                b'Content-Type: application/octet-stream\r\n',
+                b'Content-Disposition: form-data; name="files"; filename="',
+                path_escaped, b'"\r\n',
+                b'\r\n',
+            ])
+
+            content = source.read(self._chunk_size)
+            while content:
+                if isinstance(content, str):
+                    content = content.encode(encoding)
+                yield content
+                content = source.read(self._chunk_size)
+
+            yield b''.join([
+                b'\r\n',
+                b'--', boundary, b'--\r\n',
+            ])
+
+        return generator(), content_type
 
     def list_files(self, path: str, *, pattern: str = None,
                    itself: bool = False) -> typing.List[FileInfo]:
@@ -1662,7 +1841,11 @@ class Client:
 
         Args:
             path: Path of the file or directory to delete from the remote system.
-            recursive: If True, recursively delete path and everything under it.
+            recursive: If True, and path is a directory recursively deletes it and
+                       everything under it. If the path is a file, delete the file and
+                       do nothing if the file is non-existent. Behaviourally similar
+                       to `rm -rf <file|dir>`
+
         """
         info = {'path': path}
         if recursive:
@@ -1953,3 +2136,262 @@ class Client:
             'services': services,
         }
         self._request('POST', '/v1/signals', body=body)
+
+    def get_checks(
+        self,
+        level: CheckLevel = None,
+        names: typing.List[str] = None
+    ) -> typing.List[CheckInfo]:
+        """Get the check status for the configured checks.
+
+        Args:
+            level: Optional check level to query for (default is to fetch
+                checks with any level).
+            names: Optional list of check names to query for (default is to
+                fetch checks with any name).
+
+        Returns:
+            List of :class:`CheckInfo` objects.
+        """
+        query = {}
+        if level is not None:
+            query['level'] = level.value
+        if names:
+            query['names'] = names
+        resp = self._request('GET', '/v1/checks', query)
+        return [CheckInfo.from_dict(info) for info in resp['result']]
+
+
+class _FilesParser:
+    """A limited purpose multi-part parser backed by files for memory efficiency."""
+
+    def __init__(self, boundary: typing.Union[bytes, str]):
+        self._response = None
+        self._files = {}
+
+        # Prepare the MIME multipart boundary line patterns.
+        if isinstance(boundary, str):
+            boundary = boundary.encode()
+
+        # State vars, as we may enter the feed() function multiple times.
+        self._headers = None
+        self._part_type = None
+        self._response_data = bytearray()
+
+        self._max_lookahead = 8 * 1024 * 1024
+
+        self._parser = _MultipartParser(
+            boundary,
+            self._process_header,
+            self._process_body,
+            max_lookahead=self._max_lookahead)
+
+        # RFC 2046 says that the boundary string needs to be preceded by a CRLF.
+        # Unfortunately, the request library's header parsing logic strips off one of
+        # these, so we'll prime the parser buffer with that missing sequence.
+        self._parser.feed(b'\r\n')
+
+    def _process_header(self, data: bytes):
+        parser = email.parser.BytesFeedParser()
+        parser.feed(data)
+        self._headers = parser.close()
+
+        content_disposition = self._headers.get_content_disposition()
+        if content_disposition != 'form-data':
+            raise ProtocolError(
+                'unexpected content disposition: {!r}'.format(content_disposition))
+
+        name = self._headers.get_param('name', header='content-disposition')
+        if name == 'files':
+            filename = self._headers.get_filename()
+            if filename is None:
+                raise ProtocolError('multipart "files" part missing filename')
+            self._prepare_tempfile(filename)
+        elif name != 'response':
+            raise ProtocolError(
+                'unexpected name in content-disposition header: {!r}'.format(name))
+
+        self._part_type = name
+
+    def _process_body(self, data: bytes, done=False):
+        if self._part_type == 'response':
+            self._response_data.extend(data)
+            if done:
+                if len(self._response_data) > self._max_lookahead:
+                    raise ProtocolError('response end marker not found')
+                self._response = json.loads(self._response_data.decode())
+                self._response_data = bytearray()
+        elif self._part_type == 'files':
+            if done:
+                # This is the final write.
+                outfile = self._get_open_tempfile()
+                outfile.write(data)
+                outfile.close()
+                self._headers = None
+            else:
+                # Not the end of file data yet. Don't open/close file for intermediate writes
+                outfile = self._get_open_tempfile()
+                outfile.write(data)
+
+    def remove_files(self):
+        """Remove all temporary files on disk."""
+        for file in self._files.values():
+            os.unlink(file.name)
+        self._files.clear()
+
+    def feed(self, data: bytes):
+        """Provide more data to the running parser."""
+        self._parser.feed(data)
+
+    def _prepare_tempfile(self, filename):
+        tf = tempfile.NamedTemporaryFile(delete=False)
+        self._files[filename] = tf
+        self.current_filename = filename
+
+    def _get_open_tempfile(self):
+        return self._files[self.current_filename]
+
+    def get_response(self):
+        """Return the deserialized JSON object from the multipart "response" field."""
+        return self._response
+
+    def filenames(self):
+        """Return a list of filenames from the "files" parts of the response."""
+        return list(self._files.keys())
+
+    def get_file(self, path, encoding):
+        """Return an open file object containing the data."""
+        mode = 'r' if encoding else 'rb'
+        # We're using text-based file I/O purely for file encoding purposes, not for
+        # newline normalization.  newline='' serves the line endings as-is.
+        newline = '' if encoding else None
+        return open(self._files[path].name, mode, encoding=encoding, newline=newline)
+
+
+class _MultipartParser:
+    def __init__(
+            self,
+            marker: bytes,
+            handle_header,
+            handle_body,
+            max_lookahead: int = 0,
+            max_boundary_length: int = 0):
+        r"""Configures a parser for mime multipart messages.
+
+        Args:
+            marker: the multipart boundary marker (i.e. in "\r\n--<marker>--\r\n")
+
+            handle_header(data): called once with the entire contents of a part
+            header as encountered in data fed to the parser
+
+            handle_body(data, done=False): called incrementally as part body
+            data is fed into the parser - its "done" parameter is set to true when
+            the body is complete.
+
+            max_lookahead: maximum amount of bytes to buffer when searching for a complete header.
+
+            max_boundary_length: maximum number of bytes that can make up a part
+            boundary (e.g. \r\n--<marker>--\r\n")
+        """
+        self._marker = marker
+        self._handle_header = handle_header
+        self._handle_body = handle_body
+        self._max_lookahead = max_lookahead
+        self._max_boundary_length = max_boundary_length
+
+        self._buf = bytearray()
+        self._pos = 0  # current position in buf
+        self._done = False  # whether we have found the terminal boundary and are done parsing
+        self._header_terminator = b'\r\n\r\n'
+
+        # RFC 2046 notes optional "linear whitespace" (e.g. [ \t]+) after the boundary pattern
+        # and the optional "--" suffix.  The boundaries strings can be constructed as follows:
+        #
+        #     boundary = \r\n--<marker>[ \t]+\r\n
+        #     terminal_boundary = \r\n--<marker>--[ \t]+\r\n
+        #
+        # 99 is arbitrarily chosen to represent a max number of linear
+        # whitespace characters to help avoid wrongly writing boundary
+        # characters into a (temporary) file.
+        if not max_boundary_length:
+            self._max_boundary_length = len(b'\r\n--' + marker + b'--\r\n') + 99
+
+    def feed(self, data: bytes):
+        """Feeds data incrementally into the parser."""
+        if self._done:
+            return
+        self._buf.extend(data)
+
+        while True:
+            # seek to a boundary if we aren't already on one
+            i, n, self._done = _next_part_boundary(self._buf, self._marker)
+            if i == -1 or self._done:
+                return  # waiting for more data or terminal boundary reached
+
+            if self._pos == 0:
+                # parse the part header
+                if self._max_lookahead and len(self._buf) - self._pos > self._max_lookahead:
+                    raise ProtocolError('header terminator not found')
+                term_index = self._buf.find(self._header_terminator)
+                if term_index == -1:
+                    return  # waiting for more data
+
+                start = i + n
+                # data includes the double CRLF at the end of the header.
+                end = term_index + len(self._header_terminator)
+
+                self._handle_header(self._buf[start:end])
+                self._pos = end
+            else:
+                # parse the part body
+                ii, nn, self._done = _next_part_boundary(self._buf, self._marker, start=self._pos)
+                safe_bound = max(0, len(self._buf) - self._max_boundary_length)
+                if ii != -1:
+                    # part body is finished
+                    self._handle_body(self._buf[self._pos:ii], done=True)
+                    self._buf = self._buf[ii:]
+                    self._pos = 0
+                    if self._done:
+                        return  # terminal boundary reached
+                elif safe_bound > self._pos:
+                    # write partial body data
+                    data = self._buf[self._pos:safe_bound]
+                    self._pos = safe_bound
+                    self._handle_body(data)
+                    return  # waiting for more data
+                else:
+                    return  # waiting for more data
+
+
+def _next_part_boundary(buf, marker, start=0):
+    """Returns the index of the next boundary marker in buf beginning at start.
+
+    Returns:
+        (index, length, is_terminal) or (-1, -1, False) if no boundary is found.
+    """
+    prefix = b'\r\n--' + marker
+    suffix = b'\r\n'
+    terminal_midfix = b'--'
+
+    i = buf.find(prefix, start)
+    if i == -1:
+        return -1, -1, False
+
+    pos = i + len(prefix)
+    is_terminal = False
+    if buf[pos:].startswith(terminal_midfix):
+        is_terminal = True
+        pos += len(terminal_midfix)
+
+    # Note: RFC 2046 notes optional "linear whitespace" (e.g. [ \t]) after the boundary pattern
+    # and the optional "--" suffix.
+    tail = buf[pos:]
+    for c in tail:
+        if c not in b' \t':
+            break
+        pos += 1
+
+    if buf[pos:].startswith(suffix):
+        pos += len(suffix)
+        return i, pos - i, is_terminal
+    return -1, -1, False
