@@ -12,15 +12,35 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-"""Infrastructure to build unittests for Charms using the Operator Framework."""
+"""Infrastructure to build unittests for Charms using the Operator Framework.
+
+Global Variables:
+
+    SIMULATE_CAN_CONNECT: This enables can_connect simulation for the test
+    harness.  It should be set *before* you create Harness instances and not
+    changed after.  You *should* set this to true - it will help your tests be
+    more accurate!  This causes all containers' can_connect states initially
+    be False rather than True and causes the testing with the harness to model
+    and track can_connect state for containers accurately.  This means that
+    calls that require communication with the container API (e.g.
+    Container.push, Container.get_plan, Container.add_layer, etc.) will only
+    succeed if Container.can_connect() returns True and will raise exceptions
+    otherwise.  can_connect state evolves automatically to track with events
+    associated with container state, (e.g.  calling container_pebble_ready).
+    If SIMULATE_CAN_CONNECT is True, can_connect state for containers can also
+    be manually controlled using Harness.set_can_connect.
+"""
+
 
 import datetime
 import fnmatch
 import inspect
 import pathlib
 import random
+import signal
 import tempfile
 import typing
+import warnings
 from contextlib import contextmanager
 from io import BytesIO, StringIO
 from textwrap import dedent
@@ -28,6 +48,10 @@ from textwrap import dedent
 from ops import charm, framework, model, pebble, storage
 from ops._private import yaml
 
+# Toggles Container.can_connect simulation globally for all harness instances.
+# For this to work, it must be set *before* Harness instances are created.
+SIMULATE_CAN_CONNECT = False
+
 # OptionalYAML is something like metadata.yaml or actions.yaml. You can
 # pass in a file-like object or the string directly.
 OptionalYAML = typing.Optional[typing.Union[str, typing.TextIO]]
@@ -99,6 +123,23 @@ class Harness(typing.Generic[CharmType]):
         self._defaults = self._load_config_defaults(config)
         self._update_config(key_values=self._defaults)
 
+        # TODO: If/when we decide to allow breaking changes for a release,
+        # change SIMULATE_CAN_CONNECT default value to True and remove the
+        # warning message below.  This warning was added 2022-03-22
+        if not SIMULATE_CAN_CONNECT:
+            warnings.warn(
+                'Please set ops.testing.SIMULATE_CAN_CONNECT=True.'
+                'See https://juju.is/docs/sdk/testing#heading--simulate-can-connect for details.')
+
+    def set_can_connect(self, container: typing.Union[str, model.Container], val: bool):
+        """Change the simulated can_connect status of a container's underlying pebble client.
+
+        Calling this method raises an exception if SIMULATE_CAN_CONNECT is False.
+        """
+        if isinstance(container, str):
+            container = self.model.unit.get_container(container)
+        self._backend._set_can_connect(container._pebble, val)
+
     @property
     def charm(self) -> CharmType:
         """Return the instance of the charm class that was passed to __init__.
@@ -149,13 +190,18 @@ class Harness(typing.Generic[CharmType]):
         """Called when you want the Harness to fire the same hooks that Juju would fire at startup.
 
         This triggers install, relation-created, config-changed, start, and any relation-joined
-        hooks. Based on what relations have been defined before you called begin().
-        Note that all of these are fired before returning control to the test suite, so if you
-        want to introspect what happens at each step, you need to fire them directly
-        (eg Charm.on.install.emit()).
+        hooks based on what relations have been defined+added before you called begin. This does
+        NOT trigger a pebble-ready hook. Note that all of these are fired before returning control
+        to the test suite, so if you want to introspect what happens at each step, you need to fire
+        them directly (e.g. Charm.on.install.emit()).  In your hook callback functions, you should
+        not assume that workload containers are active; guard such code with checks to
+        Container.can_connect().  You are encouraged to test this by setting the global
+        SIMULATE_CAN_CONNECT variable to True.
 
         To use this with all the normal hooks, you should instantiate the harness, setup any
-        relations that you want active when the charm starts, and then call this method.
+        relations that you want active when the charm starts, and then call this method.  This
+        method will automatically create and add peer relations that are specified in
+        metadata.yaml.
 
         Example::
 
@@ -175,11 +221,12 @@ class Harness(typing.Generic[CharmType]):
             # To be fired.
         """
         self.begin()
+
         # Checking if disks have been added
         # storage-attached events happen before install
         for storage_name in self._meta.storages:
-            storage_name = storage_name.replace('-', '_')
             for storage_index in self._backend.storage_list(storage_name):
+                storage_name = storage_name.replace('-', '_')
                 # Storage device(s) detected, emit storage-attached event(s)
                 self._charm.on[storage_name].storage_attached.emit(
                     model.Storage(storage_name, storage_index, self._backend))
@@ -214,6 +261,11 @@ class Harness(typing.Generic[CharmType]):
             self._charm.on.leader_settings_changed.emit()
         self._charm.on.config_changed.emit()
         self._charm.on.start.emit()
+        # If the initial hooks do not set a unit status, the Juju controller will switch
+        # the unit status from "Maintenance" to "Unknown". See gh#726
+        post_setup_sts = self._backend.status_get()
+        if post_setup_sts.get("status") == "maintenance" and not post_setup_sts.get("message"):
+            self._backend.status_set("unknown", "", is_app=False)
         all_ids = list(self._backend._relation_names.items())
         random.shuffle(all_ids)
         for rel_id, rel_name in all_ids:
@@ -484,6 +536,11 @@ class Harness(typing.Generic[CharmType]):
     def add_relation(self, relation_name: str, remote_app: str) -> int:
         """Declare that there is a new relation between this app and `remote_app`.
 
+        In the case of adding peer relations, `remote_app` is *this* app.  This function creates a
+        relation with an application and will trigger a relation-created hook. To relate units (and
+        trigger relation-joined and relation-changed hooks), you should also call
+        :meth:`.add_relation_unit`.
+
         Args:
             relation_name: The relation on Charm that is being related to
             remote_app: The name of the application that is being related to
@@ -708,11 +765,16 @@ class Harness(typing.Generic[CharmType]):
     def container_pebble_ready(self, container_name: str):
         """Fire the pebble_ready hook for the associated container.
 
-        This will do nothing if the begin() has not been called.
+        This will do nothing if the begin() has not been called.  If
+        SIMULATE_CAN_CONNECT is True, this will switch the given
+        container's can_connect state to True before the hook
+        function is called.
         """
         if self.charm is None:
             return
         container = self.model.unit.get_container(container_name)
+        if SIMULATE_CAN_CONNECT:
+            self.set_can_connect(container, True)
         self.charm.on[container_name].pebble_ready.emit(container)
 
     def get_workload_version(self) -> str:
@@ -780,13 +842,24 @@ class Harness(typing.Generic[CharmType]):
             rel_data._invalidate()
 
         new_values = self._backend._relation_data[relation_id][app_or_unit].copy()
+        values_have_changed = False
         for k, v in key_values.items():
             if v == '':
-                new_values.pop(k, None)
+                if new_values.pop(k, None) != v:
+                    values_have_changed = True
             else:
-                new_values[k] = v
+                if k not in new_values or new_values[k] != v:
+                    new_values[k] = v
+                    values_have_changed = True
+
+        # Update the relation data in any case to avoid spurious references
+        # by an test to an updated value to be invalidated by a lack of assignment
         self._backend._relation_data[relation_id][app_or_unit] = new_values
 
+        if not values_have_changed:
+            # Do not issue a relation changed event if the data bags have not changed
+            return
+
         if app_or_unit == self._model.unit.name:
             # No events for our own unit
             return
@@ -1028,8 +1101,22 @@ class _TestingModelBackend:
         # {socket_path : _TestingPebbleClient}
         # socket_path = '/charm/containers/{container_name}/pebble.socket'
         self._pebble_clients = {}  # type: {str: _TestingPebbleClient}
+        self._pebble_clients_can_connect = {}  # type: {_TestingPebbleClient: bool}
         self._planned_units = None
 
+    def _can_connect(self, pebble_client) -> bool:
+        """Returns whether the mock client is active and can support API calls with no errors."""
+        return self._pebble_clients_can_connect[pebble_client]
+
+    def _set_can_connect(self, pebble_client, val):
+        """Manually sets the can_connect state for the given mock client."""
+        if not SIMULATE_CAN_CONNECT:
+            raise RuntimeError('must set SIMULATE_CAN_CONNECT=True before using set_can_connect')
+        if pebble_client not in self._pebble_clients_can_connect:
+            msg = 'cannot set can_connect for the client - are you running a "real" pebble test?'
+            raise RuntimeError(msg)
+        self._pebble_clients_can_connect[pebble_client] = val
+
     def _cleanup(self):
         if self._resource_dir is not None:
             self._resource_dir.cleanup()
@@ -1218,6 +1305,7 @@ class _TestingModelBackend:
         if client is None:
             client = _TestingPebbleClient(self)
             self._pebble_clients[socket_path] = client
+        self._pebble_clients_can_connect[client] = not SIMULATE_CAN_CONNECT
         return client
 
     def planned_units(self):
@@ -1259,9 +1347,15 @@ class _TestingPebbleClient:
         # Has a service been started/stopped?
         self._service_status = {}
         self._fs = _MockFilesystem()
+        self._backend = backend
+
+    def _check_connection(self):
+        if not self._backend._can_connect(self):
+            raise pebble.ConnectionError('cannot connect to pebble')
 
     def get_system_info(self) -> pebble.SystemInfo:
-        return pebble.SystemInfo(version="1.0.0")
+        self._check_connection()
+        return pebble.SystemInfo(version='1.0.0')
 
     def get_warnings(
             self, select: pebble.WarningState = pebble.WarningState.PENDING,
@@ -1283,6 +1377,7 @@ class _TestingPebbleClient:
         raise NotImplementedError(self.abort_change)
 
     def autostart_services(self, timeout: float = 30.0, delay: float = 0.1) -> pebble.ChangeID:
+        self._check_connection()
         for name, service in self._render_services().items():
             # TODO: jam 2021-04-20 This feels awkward that Service.startup might be a string or
             #  might be an enum. Probably should make Service.startup a property rather than an
@@ -1306,6 +1401,8 @@ class _TestingPebbleClient:
             raise TypeError('start_services should take a list of names, not just "{}"'.format(
                 services))
 
+        self._check_connection()
+
         # Note: jam 2021-04-20 We don't implement ChangeID, but the default caller of this is
         # Container.start() which currently ignores the return value
         known_services = self._render_services()
@@ -1337,6 +1434,9 @@ cannot perform the following tasks:
         if isinstance(services, str):
             raise TypeError('stop_services should take a list of names, not just "{}"'.format(
                 services))
+
+        self._check_connection()
+
         # TODO: handle invalid names
         # Note: jam 2021-04-20 We don't implement ChangeID, but the default caller of this is
         # Container.stop() which currently ignores the return value
@@ -1365,6 +1465,9 @@ ChangeError: cannot perform the following tasks:
         if isinstance(services, str):
             raise TypeError('restart_services should take a list of names, not just "{}"'.format(
                 services))
+
+        self._check_connection()
+
         # TODO: handle invalid names
         # Note: jam 2021-04-20 We don't implement ChangeID, but the default caller of this is
         # Container.restart() which currently ignores the return value
@@ -1397,6 +1500,9 @@ ChangeError: cannot perform the following tasks:
         else:
             raise TypeError('layer must be str, dict, or pebble.Layer, not {}'.format(
                 type(layer).__name__))
+
+        self._check_connection()
+
         if label in self._layers:
             # TODO: jam 2021-04-19 These should not be RuntimeErrors but should be proper error
             #  types. https://github.com/canonical/operator/issues/514
@@ -1430,6 +1536,7 @@ ChangeError: cannot perform the following tasks:
         return services
 
     def get_plan(self) -> pebble.Plan:
+        self._check_connection()
         plan = pebble.Plan('{}')
         services = self._render_services()
         if not services:
@@ -1442,6 +1549,8 @@ ChangeError: cannot perform the following tasks:
         if isinstance(names, str):
             raise TypeError('start_services should take a list of names, not just "{}"'.format(
                 names))
+
+        self._check_connection()
         services = self._render_services()
         infos = []
         if names is None:
@@ -1466,12 +1575,14 @@ ChangeError: cannot perform the following tasks:
 
     def pull(self, path: str, *, encoding: str = 'utf-8') -> typing.Union[typing.BinaryIO,
                                                                           typing.TextIO]:
+        self._check_connection()
         return self._fs.open(path, encoding=encoding)
 
     def push(
             self, path: str, source: typing.Union[bytes, str, typing.BinaryIO, typing.TextIO], *,
             encoding: str = 'utf-8', make_dirs: bool = False, permissions: int = None,
             user_id: int = None, user: str = None, group_id: int = None, group: str = None):
+        self._check_connection()
         if permissions is not None and not (0 <= permissions <= 0o777):
             raise pebble.PathError(
                 'generic-file-error',
@@ -1491,7 +1602,15 @@ ChangeError: cannot perform the following tasks:
 
     def list_files(self, path: str, *, pattern: str = None,
                    itself: bool = False) -> typing.List[pebble.FileInfo]:
-        files = [self._fs.get_path(path)]
+        self._check_connection()
+        try:
+            files = [self._fs.get_path(path)]
+        except FileNotFoundError:
+            # conform with the real pebble api
+            raise pebble.APIError(
+                body={}, code=404, status='Not Found',
+                message="stat {}: no such file or directory".format(path))
+
         if not itself:
             try:
                 files = self._fs.list_dir(path)
@@ -1524,6 +1643,7 @@ ChangeError: cannot perform the following tasks:
     def make_dir(
             self, path: str, *, make_parents: bool = False, permissions: int = None,
             user_id: int = None, user: str = None, group_id: int = None, group: str = None):
+        self._check_connection()
         if permissions is not None and not (0 <= permissions <= 0o777):
             raise pebble.PathError(
                 'generic-file-error',
@@ -1546,7 +1666,16 @@ ChangeError: cannot perform the following tasks:
             )
 
     def remove_path(self, path: str, *, recursive: bool = False):
-        file_or_dir = self._fs.get_path(path)
+        self._check_connection()
+        try:
+            file_or_dir = self._fs.get_path(path)
+        except FileNotFoundError:
+            if recursive:
+                # Pebble doesn't give not-found error when recursive is specified
+                return
+            raise pebble.PathError(
+                'not-found', 'remove {}: no such file or directory'.format(path))
+
         if isinstance(file_or_dir, _Directory) and len(file_or_dir) > 0 and not recursive:
             raise pebble.PathError(
                 'generic-file-error', 'cannot remove non-empty directory without recursive=True')
@@ -1555,8 +1684,46 @@ ChangeError: cannot perform the following tasks:
     def exec(self, command, **kwargs):
         raise NotImplementedError(self.exec)
 
-    def send_signal(self, sig: typing.Union[int, str], services: typing.List[str]):
-        raise NotImplementedError(self.send_signal)
+    def send_signal(self, sig: typing.Union[int, str], *service_names: str):
+        if not service_names:
+            raise TypeError('send_signal expected at least 1 service name, got 0')
+        self._check_connection()
+
+        # Convert signal to str
+        if isinstance(sig, int):
+            sig = signal.Signals(sig).name
+
+        # pebble first validates the service name, and then the signal name
+
+        plan = self.get_plan()
+        for service in service_names:
+            if service not in plan.services or not self.get_services([service])[0].is_running():
+                # conform with the real pebble api
+                message = 'cannot send signal to "{}": service is not running'.format(service)
+                body = {'type': 'error', 'status-code': 500, 'status': 'Internal Server Error',
+                        'result': {'message': message}}
+                raise pebble.APIError(
+                    body=body, code=500, status='Internal Server Error', message=message
+                )
+
+        # Check if signal name is valid
+        try:
+            signal.Signals[sig]
+        except KeyError:
+            # conform with the real pebble api
+            message = 'cannot send signal to "{}": invalid signal name "{}"'.format(
+                service_names[0],
+                sig)
+            body = {'type': 'error', 'status-code': 500, 'status': 'Internal Server Error',
+                    'result': {'message': message}}
+            raise pebble.APIError(
+                body=body,
+                code=500,
+                status='Internal Server Error',
+                message=message)
+
+    def get_checks(self, level=None, names=None):
+        raise NotImplementedError(self.get_checks)
 
 
 class NonAbsolutePathError(Exception):
